/* 
 * Send out SV packet continuously through the Ethernet port interface (Data-link)
 * 
 * This code is based on the the code on Chaitu's blog:
 * "http://linux-spawn.blogspot.se/2009/10/sending-arp-request-in-linux-using-c.html"
 * 
 * To run:  Linux machine and run the following commands
 *			gcc -o sv sv.c -lm
 *			sudo chown root:root sv
 *			sudo chmod u+s sv
 *			./sv <Ethernet port name>
 *			for example: ./sv eth0
*/

#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h> /* the L2 protocols */
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <stdlib.h>/*for exit*/
#include <ifaddrs.h>/*for getifaddr*/
#include <sys/types.h>
#include <sys/ioctl.h>/*for netdevice*/
#include <net/if.h>
#include <sys/time.h>/*for time*/
#include <math.h>/*for sine wave*/

void ErrorHandling(char *e){
	perror(e);
	exit(1);
}

static int SmpCount= 0x0000;

int print_pkt(char *buf, int len){
	int j = 0;
	for(j = 0; j < len; j++ )
	{
		if((j%16) == 0 && j != 0 )
		printf("\n");
		printf("%02x ",*(buf+j)& 0xFF );
	}
	printf("\n====Packet END====\n");
	return 0;
} 

/*====START====*/
int main(int argc, char *argv[]){

	struct sockaddr_ll dsti;
	struct sockaddr_in *hostAddr, TrgtIP, SndIP;
	int packet_socket,sock,i=0;
	int8_t *buffer;
	int32_t l;
	struct ifreq ifr;

	/*====MSVCB02 Dataset PhsMeas1====*/
	struct PhsMeas1 {
		/*Current Phase A, B, C and neutral instant magnitude in Ampere*/
		int32_t CurrentATCTR1_Amp_instMag_i;
		int32_t CurrentATCTR1_Amp_q;//quality
		int32_t CurrentBTCTR2_Amp_instMag_i;
		int32_t CurrentBTCTR2_Amp_q;
		int32_t CurrentCTCTR3_Amp_instMag_i;
		int32_t CurrentCTCTR3_Amp_q;
		int32_t CurrentNTCTR4_Amp_instMag_i;
		int32_t CurrentNTCTR4_Amp_q;
		/*Voltage Phase A, B ,C and neutral instant magnitude in Volts*/
		int32_t VoltageATVTR1_Vol_instMag_i;
		int32_t VoltageATVTR1_Vol_q;/*quality*/
		int32_t VoltageBTVTR2_Vol_instMag_i;
		int32_t VoltageBTVTR2_Vol_q;
		int32_t VoltageCTVTR3_Vol_instMag_i;
		int32_t VoltageCTVTR3_Vol_q;
		int32_t VoltageNTVTR4_Vol_instMag_i;
		int32_t VoltageNTVTR4_Vol_q;
	}__attribute__((__packed__));

	/*====ASDU Application Service Data Unit====*/
	/*ASDU according to 9-2 LE and ASN.1 encoding rules*/
	struct ASDU_LE {
		int8_t SequenceASDU_tag;/*0x30*/
		int8_t SequenceASDU_length;/*0x01*/
		int8_t svID_tag;/*0x80*/
		int8_t svID_length;/*0x0A*/
		int8_t svID[11];/*naming rules is in IEC 61869-9*/
		int8_t smpCnt_tag;/*0x82*/
		int8_t smpCnt_length;/*0x02*/
		int16_t smpCnt;/*Counter specification see IEC 60044-8*/
		int8_t confRev_tag;/*0x83*/
		int8_t confRev_length;/*0x04*/
		int32_t confRev;/*Configuration revision number*/
		int8_t smpSynch_tag;/*0x85*/
		int8_t smpSynch_length;/*0x01*/
		int8_t smpSynch;
		//int8_t smpRate_tag;/*0x86*/
		//int8_t smpRate_length;/*0x02*/
		//int16_t smpRate;/*0x0050*/
		int8_t SequenceofData_tag;/*0x87*/
		int8_t SequenceofData_length;/*0x40*/
		struct PhsMeas1 DataSet;
	}__attribute__((__packed__));

	/*====APDU Application Protocol Data Unit====*/
	struct APDU {
		int8_t savPDU_tag;
		int8_t savPDU_length;/*sizeof APDU*/
		int8_t noASDU_tag;
		int8_t noASDU_length;/*sizeof noASDU*/
		int8_t noASDU;
		int8_t SequenceofASDU_tag;
		int8_t SequenceofASDU_length;/*sizeof all ASDU*/
		struct ASDU_LE ASDU1;
	}__attribute__((__packed__));

	/*====ISO/IEC 8802-3 frame structure====*/
	struct FRM_8802 {
		/*Header MAC*/
		int8_t dest_mac[6];
		int8_t src_mac[6];
		/*Priority tagged*/
		//int16_t TPID;//0x8100 for 802.1Q Virtual LAN
		//int16_t TCI;//0x4000 see 9-2 LE
		/*Ethertype PDU*/
		int16_t Ethertype;/*0x88BA for Sampled Value*/
		int16_t APPID;/*0x4000 application identifier*/
		int16_t lgth; /*start from APPID*/
		int16_t rsrv1;
		int16_t rsrv2;
		/*APDU application layer protocol data unit*/
		struct APDU apdu;
		/*pad bytes if necessary*/
		/*Frame check sequence*/
	}__attribute__((__packed__));

	struct FRM_8802 FRM;
	memset(&FRM, 0, sizeof FRM);

	printf("====Initialize parameters succeess====\n");

	/*check input argument*/
	if (argc!=2)
		ErrorHandling("To run: ./sv + <host interface name>\n");

	/*open a socket for ioctl*/
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock == -1)
		ErrorHandling("open socket for interface");
	memcpy(ifr.ifr_name, argv[1], IF_NAMESIZE);

	/*get MAC address of the interface, after call this hostAssr will be reset?*/
	if ((ioctl(sock, SIOCGIFHWADDR, &ifr, sizeof ifr)) == -1)
		ErrorHandling("HW addr");
	/*display*/
	printf("Sender MAC address: %02x:%02x:%02x:%02x:%02x:%02x \n",
								ifr.ifr_hwaddr.sa_data[0]&0xFF,
								ifr.ifr_hwaddr.sa_data[1]&0xFF,
								ifr.ifr_hwaddr.sa_data[2]&0xFF,
								ifr.ifr_hwaddr.sa_data[3]&0xFF,
								ifr.ifr_hwaddr.sa_data[4]&0xFF,
								ifr.ifr_hwaddr.sa_data[5]&0xFF);

	printf("====Format packet====\n");	
	/*====format the packet====*/
	/*MAC Header*/
	memset(FRM.dest_mac, 0xBC, sizeof FRM.dest_mac[0]);	/*Destination MAC Address, should be in the range of 01:0C:CD:04:XX:XX*/
	memset(FRM.dest_mac+1, 0x30, sizeof FRM.dest_mac[0]);
	memset(FRM.dest_mac+2, 0x5B, sizeof FRM.dest_mac[0]);
	memset(FRM.dest_mac+3, 0xA1, sizeof FRM.dest_mac[0]);
	memset(FRM.dest_mac+4, 0x7A, sizeof FRM.dest_mac[0]);
	memset(FRM.dest_mac+5, 0x1C, sizeof FRM.dest_mac[0]);
	memset(FRM.src_mac, (ifr.ifr_hwaddr.sa_data[0]&0xFF), sizeof FRM.src_mac[0]);	/*Source MAC Address*/
	memset(FRM.src_mac+1, (ifr.ifr_hwaddr.sa_data[1]&0xFF), sizeof FRM.src_mac[0]);
	memset(FRM.src_mac+2, (ifr.ifr_hwaddr.sa_data[2]&0xFF), sizeof FRM.src_mac[0]);
	memset(FRM.src_mac+3, (ifr.ifr_hwaddr.sa_data[3]&0xFF), sizeof FRM.src_mac[0]);
	memset(FRM.src_mac+4, (ifr.ifr_hwaddr.sa_data[4]&0xFF), sizeof FRM.src_mac[0]);
	memset(FRM.src_mac+5, (ifr.ifr_hwaddr.sa_data[5]&0xFF), sizeof FRM.src_mac[0]);
	//Priority tag
	//FRM.TPID = htons(0x8100);
	//FRM.TCI = htons(0x4000);
	//Ethertype PDU
	FRM.Ethertype = htons(0x88BA);
	FRM.APPID = htons(0x4000);
	FRM.lgth = htons(0x006D);
	FRM.rsrv1 = htons(0x0000);/*htons(0x8000);*/
	FRM.rsrv2 = htons(0x0000);
	/*APDU*/
	FRM.apdu.savPDU_tag = 0x60;
	FRM.apdu.savPDU_length = 0x63;
	FRM.apdu.noASDU_tag = 0x80;
	FRM.apdu.noASDU_length = 0x01;
	FRM.apdu.noASDU = 0x01;
	FRM.apdu.SequenceofASDU_tag = 0xA2;
	FRM.apdu.SequenceofASDU_length = 0x5E;
	/*ASDU1*/
	FRM.apdu.ASDU1.SequenceASDU_tag = 0x30;
	FRM.apdu.ASDU1.SequenceASDU_length = 0x5C;
	FRM.apdu.ASDU1.svID_tag = 0x80;
	FRM.apdu.ASDU1.svID_length = 0x0B;
	/*ASDU1*/
	FRM.apdu.ASDU1.SequenceASDU_tag = 0x30;
	FRM.apdu.ASDU1.SequenceASDU_length = 0x5C;
	FRM.apdu.ASDU1.svID_tag = 0x80;
	FRM.apdu.ASDU1.svID_length = 0x0B;
	memset(FRM.apdu.ASDU1.svID, 0x5F, sizeof FRM.apdu.ASDU1.svID[0]);/*_*/
	memset(FRM.apdu.ASDU1.svID+1, 0x41, sizeof FRM.apdu.ASDU1.svID[0]);/*A*/
	memset(FRM.apdu.ASDU1.svID+2, 0x48, sizeof FRM.apdu.ASDU1.svID[0]);/*H*/
	memset(FRM.apdu.ASDU1.svID+3, 0x4D, sizeof FRM.apdu.ASDU1.svID[0]);/*M*/
	memset(FRM.apdu.ASDU1.svID+4, 0x45, sizeof FRM.apdu.ASDU1.svID[0]);/*E*/
	memset(FRM.apdu.ASDU1.svID+5, 0x44, sizeof FRM.apdu.ASDU1.svID[0]);/*D*/
	memset(FRM.apdu.ASDU1.svID+6, 0x5F, sizeof FRM.apdu.ASDU1.svID[0]);/*_*/
	memset(FRM.apdu.ASDU1.svID+7, 0x53, sizeof FRM.apdu.ASDU1.svID[0]);/*S*/
	memset(FRM.apdu.ASDU1.svID+8, 0x56, sizeof FRM.apdu.ASDU1.svID[0]);/*V*/
	memset(FRM.apdu.ASDU1.svID+9, 0x31, sizeof FRM.apdu.ASDU1.svID[0]);/*1*/
	memset(FRM.apdu.ASDU1.svID+10, 0x5F, sizeof FRM.apdu.ASDU1.svID[0]);/*_*/
	FRM.apdu.ASDU1.smpCnt_tag = 0x82;
	FRM.apdu.ASDU1.smpCnt_length = 0x02;
	FRM.apdu.ASDU1.confRev_tag = 0x83;
	FRM.apdu.ASDU1.confRev_length = 0x04;
	FRM.apdu.ASDU1.confRev = htonl(0x00000001);
	FRM.apdu.ASDU1.smpSynch_tag = 0x85;
	FRM.apdu.ASDU1.smpSynch_length = 0x01;
	FRM.apdu.ASDU1.smpSynch = 0x00;/*0-none;1-local;2-global*/
	FRM.apdu.ASDU1.SequenceofData_tag = 0x87;
	FRM.apdu.ASDU1.SequenceofData_length = 0x40;
	/*====format END====*/

	printf("====Format success====\n");
	print_pkt((char *)&FRM, sizeof FRM);

	/*get interface index*/
	if ((ioctl(sock, SIOCGIFINDEX, &ifr, sizeof ifr)) == -1)
		ErrorHandling("index");
		
	/*set the destination sockaddr_ll*/
	memset(&dsti, 0, sizeof dsti);
	dsti.sll_family=AF_PACKET;
	dsti.sll_protocol=htons(ETH_P_8021Q); /*it also works with ETH_P_ARP*/
	dsti.sll_ifindex=ifr.ifr_ifindex;

	/*open a layer 2 socket*/
	packet_socket = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_8021Q));
	if(packet_socket == -1)
		ErrorHandling("open layer 2 socket");
	else printf("open layer 2 socket succeed\nsending...\n");
	
	/*send ARP to dsti*/
	int cnt=0,fault=6; 
	/*the SV will be sent out continuously*/
	while(1){
	struct timeval tv1,tv2;
	struct timezone tz1,tz2;
	int def=0,slp=0,cmpst=0;
	int Ia,Ib,Ic,Va,Vb,Vc;
	double CurrentA,VoltageA,CurrentB,VoltageB,CurrentC,VoltageC;
	
		for (i=0;i<=3999;i++){
			if (SmpCount==0x0FA0){/*0x3200*/
				SmpCount=0x0000;
				cnt++;
			}
			FRM.apdu.ASDU1.smpCnt = htons(SmpCount);
			SmpCount++;
			if (cnt==15){/*fault current will appear after 15 seconds*/
				fault=1000000;
				
			}
			if (cnt==20){/*fault current will be removed after 20 seconds*/
				fault=6;
				
			}
			/*Fake sinusoidal values*/
			CurrentA = fault*sin(100*3.1415926*i*1/4000)*1000;/*the current scale factor for SV is 0.001*/
			CurrentB = fault*sin(100*3.1415926*i*1/4000+3.1415926*2/3)*1000;
			CurrentC = fault*sin(100*3.1415926*i*1/4000-3.1415926*2/3)*1000;
			VoltageA = 100000*sin(100*3.1415926*i*1/4000)*100;/*the voltage scale factor for SV is 0.01*/
			VoltageB = 100000*sin(100*3.1415926*i*1/4000+3.1415926*2/3)*100;
			VoltageC = 100000*sin(100*3.1415926*i*1/4000-3.1415926*2/3)*100;
			/*Assign values into PhsMeas1*/
			FRM.apdu.ASDU1.DataSet.CurrentATCTR1_Amp_instMag_i = htonl(CurrentA);
			FRM.apdu.ASDU1.DataSet.CurrentATCTR1_Amp_q = htonl(0x00000000);/*for detailed identifier description see the thesis report of the author*/
			FRM.apdu.ASDU1.DataSet.CurrentBTCTR2_Amp_instMag_i = htonl(CurrentB);
			FRM.apdu.ASDU1.DataSet.CurrentBTCTR2_Amp_q = htonl(0x00000000);
			FRM.apdu.ASDU1.DataSet.CurrentCTCTR3_Amp_instMag_i = htonl(CurrentC);
			FRM.apdu.ASDU1.DataSet.CurrentCTCTR3_Amp_q = htonl(0x00000000);
			FRM.apdu.ASDU1.DataSet.CurrentNTCTR4_Amp_instMag_i = htonl(CurrentA+CurrentB+CurrentC);/*neutral*/
			FRM.apdu.ASDU1.DataSet.CurrentNTCTR4_Amp_q = htonl(0x00002000);
			FRM.apdu.ASDU1.DataSet.VoltageATVTR1_Vol_instMag_i = htonl(VoltageA);
			FRM.apdu.ASDU1.DataSet.VoltageATVTR1_Vol_q = htonl(0x00000000);
			FRM.apdu.ASDU1.DataSet.VoltageBTVTR2_Vol_instMag_i = htonl(VoltageB);
			FRM.apdu.ASDU1.DataSet.VoltageBTVTR2_Vol_q = htonl(0x00000000);
			FRM.apdu.ASDU1.DataSet.VoltageCTVTR3_Vol_instMag_i = htonl(VoltageC);
			FRM.apdu.ASDU1.DataSet.VoltageCTVTR3_Vol_q = htonl(0x00000000);
			FRM.apdu.ASDU1.DataSet.VoltageNTVTR4_Vol_instMag_i = htonl(VoltageA+VoltageB+VoltageC);/*neutral*/
			FRM.apdu.ASDU1.DataSet.VoltageNTVTR4_Vol_q = htonl(0x00002000);
			if(i == 0){/*get the locol time of the computer when SV starts*/
				gettimeofday(&tv1, &tz1);
			}
			if(sendto(packet_socket, &FRM, sizeof FRM, 0, (struct sockaddr *)&dsti, sizeof dsti)==-1)
				ErrorHandling("sendto()");
				
			gettimeofday(&tv2, &tz2);/*get after each time when SV is sent out*/
			def=1000000/4000*i-((tv2.tv_sec-tv1.tv_sec)*1000000+(tv2.tv_usec-tv1.tv_usec))+cmpst;/*calculate the witing time*/
			if(def<0){
				slp = 0;
				cmpst = def;
			}else{
				slp = def;
				cmpst = 0;
			}
			usleep(slp);/*suspend thread to ensure the next SV is sent out 250us after the previous SV*/
		}
	}
	printf("====Packet sent====\n");
	print_pkt((char *)&FRM, sizeof FRM);
	/*close sockets*/
	close(packet_socket);
	close(sock);
}
